""" parser functions """
import optparse


def add_entry_parser() -> optparse.OptionParser:
    """ Parser for entry command. """
    parser = optparse.OptionParser(
        usage="main.py add-entry [options]\n" "Add pair to to the list"
    )
    parser.add_option(
        "-s", "--stock", dest="stock", help="Symbol of the stock to entry."
    )
    parser.add_option(
        "-d", "--date", dest="date", help="Date of the entry, please add as DD/MM/YYYY"
    )
    parser.add_option("-c", "--count", dest="count", help="Count of the stocks bought.")
    parser.add_option("-p", "--price", dest="price", help="Price of the stocks.")
    return parser

def import_parser() -> optparse.OptionParser:
    """ Parser for import command """
    parser = optparse.OptionParser(
        usage="main.py import-data [options]\n" "Import data from your custom files."
    )
    parser.add_option(
        "-e", "--entries", dest="entries", help="Path to your custom file with entries."
    )
    parser.add_option(
        "-p", "--portfolio", dest="portfolio", help="Path to your custom file with portfolio."
    )

    return parser

def export_parser() -> optparse.OptionParser:
    """ Parser for export command."""
    parser = optparse.OptionParser(
        usage="main.py import-data [options]\n" "Export data from your custom files to some"
              "directory."
    )
    parser.add_option(
        "-d", "--directory", dest="directory", help="Path to directory for the output."
    )

    return parser



""" Module with main functionality of the tool"""
import csv
import logging
import os
import datetime
import pathlib
import shutil
import platform
import sys
import webbrowser
from typing import List

import jinja2
logging_level = os.environ.get("DEBUG_LEVEL")
logging.basicConfig(level=logging_level if logging_level else "INFO")
from stock_summary.library import (export_data, get_entries_summary,
                                   get_exchange_rates, get_pair_prices,
                                   get_plot_html, import_data,
                                   prepare_portfolio_data, validate_date)
from stock_summary.parsers import (add_entry_parser, export_parser,
                                   import_parser)
from stock_summary.settings import (ENTRIES_PATH, INDEX_HTML_FILE,
                                    MAIN_CSS_FILE, PORTFOLIO_PATH,
                                    TOKEN_PATH, SETTINGS_PATH)







def get_pairs() -> List[str]:
    """Returns list of pairs."""
    with open(ENTRIES_PATH, newline="", encoding="utf-8") as csvfile:
        pair_lines = csv.reader(csvfile, delimiter=" ", quotechar="|")
        next(pair_lines)
        pairs = [pair[1].strip() for pair in pair_lines if pair]
        logging.debug("Getting pairs %s", pairs)
        return list(set(pairs))


def generate_portfolio_main() -> None:
    """Generates portfolio for actual time and entered entries."""
    conversion_rates = get_exchange_rates()
    pairs = get_pairs()
    prices = get_pair_prices(pairs)
    with open(ENTRIES_PATH, newline="", encoding="utf-8") as csvfile:
        entries_reader = csv.reader(csvfile, delimiter=" ", quotechar="|")
        init_value: float = 0
        curr_value: float = 0
        next(entries_reader)
        for row in entries_reader:
            if not row:
                continue
            count = float(row[2])
            init_price = float(row[3])
            conversion_rate = conversion_rates[prices[row[1]]["currency"]]
            init_value += count * init_price * conversion_rate
            curr_value += count * prices[row[1]]["regularMarketPrice"] * conversion_rate
    with open(PORTFOLIO_PATH, "a", encoding="utf-8") as result_file:
        now = datetime.datetime.now()
        result_file.write(
            f"{now.strftime('%d/%m/%y')} {curr_value} {(curr_value / init_value - 1) * 100}\n"
        )
    logging.info(
        "Portfolio with price %s and percentage move %s generated and added.",
        curr_value,
        (curr_value / init_value - 1) * 100,
    )

def save_entry(date: str, stock: str, count: str, price: str) -> None:
    """Save entries to CSV file."""
    with open(ENTRIES_PATH, "a", newline="", encoding="utf-8") as csvfile:
        csv_writer = csv.writer(csvfile, delimiter=" ", quotechar="|")
        csv_writer.writerow([date, stock, count, price])
    logging.debug(f"Entry date: {date} stock: {stock} count: {count} price: {price} saved to {ENTRIES_PATH}")

def generate_html_main() -> None:
    """
    Generates HTML and executes it in your browser.
    """
    summary_records = list(get_entries_summary().values())
    portfolio_data = prepare_portfolio_data()
    plot_html = get_plot_html(portfolio_data)
    environment = jinja2.Environment()
    with open(f"{pathlib.Path(__file__).parent.resolve()}/html_files/jinja.html", "r", encoding="utf-8") as html_template:
        template = environment.from_string(html_template.read())
    with open(INDEX_HTML_FILE, "w", encoding="utf-8") as index_file:
        index_file.write(template.render(plot_html=plot_html, records=summary_records))
    shutil.copy2(f"{pathlib.Path(__file__).parent.resolve()}/html_files/main.css", MAIN_CSS_FILE)
    logging.info(f"Index html file successfully saved to {INDEX_HTML_FILE}")
    prepend_str = ""
    if platform.system().lower() == "darwin":
        prepend_str = "file:///"
    webbrowser.open(f"{prepend_str}{INDEX_HTML_FILE}")


def add_entry_main() -> None:
    """
    Root function for the add entry argument from the command line.
    """
    parser = add_entry_parser()
    (options, _) = parser.parse_args()
    if not options.stock or not options.date or not options.count or not options.price:
        logging.error("You have to enter all needed params")
        raise ValueError("You have to enter all needed params")
    pair = options.stock.strip()
    if pair not in get_pairs():
        raise ValueError("Adding non existing stock symbol.")
    try:
        validate_date(options.date)
        int(options.count)
        int(options.price)
    except TypeError as err:
        raise RuntimeError("parameters have bad types, please try again") from err
    save_entry(options.date, options.stock, options.count, options.price)
    logging.info(
        "Entry with date %s , stock %s, count %s , price %s successfully added.",
        options.date,
        options.stock,
        options.count,
        options.price,
    )

def export_data_main() -> None:
    """ Main function for export data command."""
    parser = export_parser()
    (options, _) = parser.parse_args()
    if not options.directory:
        logging.error("You need to pass the directory for the output.")
        sys.exit(1)
    export_data(options.directory)

def import_data_main() -> None:
    """ Main function for import data command."""
    parser = import_parser()
    (options, _) = parser.parse_args()
    if options.portfolio:
        import_data(options.portfolio, PORTFOLIO_PATH)
        logging.info(f"Portfolio {options.portfolio} successfully updated to {PORTFOLIO_PATH}")
    if options.entries:
        import_data(options.entries, ENTRIES_PATH)
        logging.info(f"Entries {options.entries} successfully updated to {ENTRIES_PATH}")

def save_token_main() -> None:
    """ Main function for save token command."""
    token = sys.argv[2]
    os.makedirs(SETTINGS_PATH, exist_ok=True)
    with open(TOKEN_PATH, "w", encoding="utf-8") as token_file:
        token_file.write(token)
    logging.info(f"Token successfully saved to {TOKEN_PATH}")


def print_main_help() -> None:
    """ Prints main help """
    print(
        "Available commands:\n\n"
        "add-entry [opts] - add entry for current date, stock, count and price\n\n"
        "generate-portfolio [opts] - generates actual value of your portfolio in CZK and "
        "percentage move from the start of your investments\n"
        "generate-html [opts] - generates summary page from all current data\n\n"
        "import-data [opts] - imports data from your custom files\n\n"
        "export-data [opts] - exports data to your directory\n\n"
        "save-token (token) - save your token to rapidAPI\n\n"
    )

def main() -> None:
    """ Main function."""
    try:
        if sys.argv[1] == "generate-html":
            generate_html_main()
        elif sys.argv[1] == "add-entry":
            add_entry_main()
        elif sys.argv[1] == "generate-portfolio":
            generate_portfolio_main()
        elif sys.argv[1] == "export-data":
            export_data_main()
        elif sys.argv[1] == "import-data":
            import_data_main()
        elif sys.argv[1] == "save-token":
            save_token_main()
        elif sys.argv[1] == "-h" or sys.argv[1] == "--help":
            print_main_help()
        else:
            logging.error(
                "Invalid first command, use option from: 'retrain-model','make-analysis"
                ",'retrain-model','get-result'"
            )
            sys.exit(1)
    except (RuntimeError, ValueError) as err:
        logging.error(err)
        raise err
    sys.exit(1)
# Press the green button in the gutter to run the script.
if __name__ == "__main__":
    main()



""" library functions """
import csv
import datetime
import json
import os
import pathlib
import shutil
import sys
import logging
from pathlib import Path
from typing import Any, Dict, List

import pandas as pd
import plotly.express as px
import requests

from stock_summary.settings import (DATA_PATH, ENTRIES_PATH,
                                    EXCHANGE_RATE_HEADERS, EXCHANGE_RATE_URL,
                                    PORTFOLIO_PATH, STOCK_PRICE_HEADERS,
                                    STOCK_PRICE_URL,
                                    PairResponse, SummaryDict)



def check_pair_responses(pairs: List[PairResponse]) -> None:
    """Do same basic checks on incoming responses from API for stock prices."""
    for pair in pairs:
        if pair["regularMarketPrice"] <= 0:
            raise ValueError(f"Entered symbol {pair['symbol']} has invalid price <=0.")
    logging.debug(f"Successfully checked responses for {pairs}")

def get_exchange_rates() -> Dict[str, float]:
    """Returns dict with actual values for conversions between other currencies and CZK"""

    response = requests.request("GET", EXCHANGE_RATE_URL, headers=EXCHANGE_RATE_HEADERS, timeout=10)
    exchange_dict = {}
    for key, value in json.loads(response.text)["rates"].items():
        exchange_dict[key] = 1 / value
    logging.debug("Returning exchange dict %s", exchange_dict)
    return exchange_dict


def get_pair_prices(pairs: List[str]) -> Dict[str, PairResponse]:
    """
    Search for price on url and xpath, if error occurs, raises error.
    """
    url = f"{STOCK_PRICE_URL}/{','.join(pairs)}"

    response = requests.request("GET", url, headers=STOCK_PRICE_HEADERS, timeout=10)
    logging.debug("Requesting URL %s with response %s", url, response)
    result_list: List[PairResponse] = json.loads(response.text)
    check_pair_responses(result_list)
    result_dict = {}
    for result in result_list:
        result_dict[result["symbol"]] = result
    logging.debug(f"Returning pair_prices {result_dict}")
    return result_dict


def validate_date(date_text: str) -> None:
    """Validates date to our custom format."""
    try:
        datetime.datetime.strptime(date_text, "%d/%m/%Y")
    except ValueError as err:
        raise ValueError("Incorrect data format, should be DD/MM/YYYY") from err
    logging.debug(f"Validated data for {date_text}")


def get_entries_summary() -> Dict[str, SummaryDict]:
    """
    Returns entries summary with keys as stock symbols and values as dicts with all important
    values (see TypedDict) in settings.
    """
    entries_dict: Dict[str, SummaryDict] = {}
    with open(ENTRIES_PATH, newline="", encoding="utf-8") as csvfile:
        entries = csv.reader(csvfile, delimiter=" ", quotechar="|")
        next(entries)
        for entry in entries:
            if entry[1] not in entries_dict:
                entries_dict[entry[1]] = {
                    "symbol": entry[1],
                    "count": 0,
                    "cost_basis": 0,
                    "actual_basis": 0,
                    "currency": "",
                    "actual_price": 0
                }
            entries_dict[entry[1]]["count"] += float(entry[2])
            entries_dict[entry[1]]["cost_basis"] += float(entry[2]) * float(entry[3])
    exchange_rates = get_exchange_rates()
    pair_prices = get_pair_prices(list(entries_dict.keys()))
    for key, value in entries_dict.items():
        value["currency"] = pair_prices[key]["currency"]
        value["actual_price"] = pair_prices[key]["regularMarketPrice"]
        value["cost_basis"] = (
            value["cost_basis"] * exchange_rates[pair_prices[key]["currency"]]
        )
        value["actual_basis"] = (
            value["count"] * pair_prices[key]["regularMarketPrice"]
        ) * exchange_rates[pair_prices[key]["currency"]]
    logging.debug(f"Returning entries summary {entries_dict}")
    return entries_dict


def prepare_portfolio_data() -> Any:
    """ Prepares portfolio data and returns them as pandas dataset"""
    dataset = pd.read_csv(PORTFOLIO_PATH, sep=" ")
    dataset["DATE"] = pd.to_datetime(dataset["DATE"], format="%d/%m/%y")
    return dataset


def get_plot_html(dataset: Any) -> Any:
    """ Exports plot as HTML and returns it."""
    fig = px.line(
        dataset,
        x="DATE",
        y="TOTAL_PRICE",
        title="Přehled portfolia v čase",
        labels={"DATE": "Čas", "TOTAL_PRICE": "Hodnota portfolia"},
        hover_data={"DATE": "|%d/%m/%y"},
    )
    return fig.to_html()

def check_if_files_exist() -> None:
    """ Check if files exist in memory, if not then creates them with initialized headers."""
    os.makedirs(DATA_PATH, exist_ok=True)
    if not os.path.exists(ENTRIES_PATH):
        shutil.copy2(f"{pathlib.Path(__file__).parent.resolve()}/init_datasets/entries",
                     ENTRIES_PATH)
        logging.debug(f"Created init entries file {ENTRIES_PATH}")
    if not os.path.exists(PORTFOLIO_PATH):
        shutil.copy2(f"{pathlib.Path(__file__).parent.resolve()}/init_datasets/portfolio",
                     PORTFOLIO_PATH)
        logging.debug(f"Created init portfolio file {ENTRIES_PATH}")

def import_data(from_file: str, to_file: str) -> None:
    """
    Imports data from entry file to target file. If target file already exists, then
    asks user for confirmation
    """
    if os.path.exists(to_file):
        confirmation = input("You will rewrite your actual saved data, are you sure to proceed? "
                             "Type Y: ")
        if not confirmation.lower() == "y":
            logging.error("Action canceled, ending without any action.")
            sys.exit(1)
    shutil.copy2(from_file, to_file)
    logging.debug(f"Successfully moved data from {from_file} to {to_file}")

def export_data(directory: str) -> None:
    """
    Exports all data files to target directory, creates recursively, if doesn't exist.
    """
    path = Path(directory)
    os.makedirs(path, exist_ok=True)
    shutil.copy2(ENTRIES_PATH, f"{path}/{ENTRIES_PATH.split('/')[-1]}")
    shutil.copy2(PORTFOLIO_PATH, f"{path}/{PORTFOLIO_PATH.split('/')[-1]}")



""" file with settings """
import logging
from enum import Enum
from typing import TypedDict

import appdirs

DATA_PATH = appdirs.user_data_dir('stock_summary')
SETTINGS_PATH = appdirs.user_config_dir('stock_summary')
class LoggingSettings(Enum):
    """ Enum which saves settings for logging"""
    OUTPUT_FILE = "tmp.log"
    LOGGING_LEVEL = logging.DEBUG


class PairResponse(TypedDict):
    """ Pair response that we get from API for each pair."""
    symbol: str
    regularMarketPrice: float
    currency: str

class SummaryDict(TypedDict):
    """ This dict is used to summarize entries for view Jinja HTML"""
    symbol: str
    actual_price: float
    actual_basis: float
    cost_basis: float
    count: float
    currency: str




PAIR_PATH = f"{DATA_PATH}/pairs"
ENTRIES_PATH = f"{DATA_PATH}/entries"
PORTFOLIO_PATH = f"{DATA_PATH}/portfolio"
INDEX_HTML_FILE = f"{DATA_PATH}/index.html"
MAIN_CSS_FILE = f"{DATA_PATH}/main.css"
TOKEN_PATH = f"{SETTINGS_PATH}/token"
try:
    with open(TOKEN_PATH,"r", encoding="utf-8") as token_file:
        API_TOKEN = token_file.read().strip()
except FileNotFoundError:
    API_TOKEN = ""

EXCHANGE_RATE_URL = "https://exchangerate-api.p.rapidapi.com/rapid/latest/CZK"
EXCHANGE_RATE_HEADERS = {
    "X-RapidAPI-Key": API_TOKEN,
    "X-RapidAPI-Host": "exchangerate-api.p.rapidapi.com",
}

STOCK_PRICE_URL = "https://yahoo-finance15.p.rapidapi.com/api/yahoo/qu/quote"
STOCK_PRICE_HEADERS = {
    "X-RapidAPI-Key": API_TOKEN,
    "X-RapidAPI-Host": "yahoo-finance15.p.rapidapi.com",
}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['stock_summary']

package_data = \
{'': ['*'],
 'stock_summary': ['demo_datasets/*', 'html_files/*', 'init_datasets/*']}

install_requires = \
['appdirs>=1.4.4,<2.0.0',
 'jinja2>=3.1.2,<4.0.0',
 'pandas>=1.5.2,<2.0.0',
 'plotly>=5.11.0,<6.0.0',
 'requests>=2.28.1,<3.0.0']

entry_points = \
{'console_scripts': ['stock_summary_tool = stock_summary.main:main']}

setup_kwargs = {
    'name': 'stock-summary-tool',
    'version': '0.1.0',
    'description': 'Tool for tracking of your investments and your actual portfolio',
    'long_description': 'DEMO\n2. Create virtual environment that you will use for the project and activate it (python3.8+ required):\n   1. **python3 -m venv my_venv/**\n   2. **source my_venv/bin/activate**\n2. Install the package \n   1. **pip3 install stock_summary_tool**\n3. Go to https://rapidapi.com and log in.\n4. Go to https://rapidapi.com/sparior/api/yahoo-finance15/ and obtain your API key.\n5. Save your key for the project:\n   1. **stock_summary_tool save-token <YOUR_TOKEN>**\n\n\n',
    'author': 'Simon Foucek',
    'author_email': 'foucek.simon@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.12',
}


setup(**setup_kwargs)
